import typing
import datetime
import dateutil.relativedelta
import pandas
import matplotlib
import matplotlib.pyplot
import tempfile
import os
import pyxirr
from .nse_product import NSEProduct
from .core import Core


class NSETRI:

    '''
    Provides functionality for downloading and analyzing
    NSE Equity Total Return Index (TRI) data,
    including both price and dividend reinvestment.
    '''

    @property
    def _index_api(
        self
    ) -> dict[str, str]:

        '''
        Returns a dictionary containing equity indices as keys
        and corresponding API names as values.
        '''

        df = NSEProduct()._dataframe_equity_index
        output = dict(
            zip(df['Index Name'], df['API TRI'])
        )

        return output

    @property
    def non_open_source_indices(
        self
    ) -> list[str]:

        '''
        Returns a list of equity indices that are not open-source.
        '''

        df = NSEProduct()._dataframe_equity_index
        df = df[df['API TRI'] == 'NON OPEN SOURCE']
        output = list(df['Index Name'].sort_values())

        return output

    def is_index_open_source(
        self,
        index: str,
    ) -> bool:

        '''
        Check whether the index data is open-source.

        Parameters
        ----------
        index : str
            Name of the index.

        Returns
        -------
        bool
            True if the index data is open-source, False otherwise.
        '''

        if NSEProduct().is_index_exist(index):
            pass
        else:
            raise Exception(f'"{index}" index does not exist.')

        output = index not in self.non_open_source_indices

        return output

    def download_historical_daily_data(
        self,
        index: str,
        excel_file: str,
        start_date: typing.Optional[str] = None,
        end_date: typing.Optional[str] = None,
        http_headers: typing.Optional[dict[str, str]] = None
    ) -> pandas.DataFrame:

        '''
        Downloads historical daily closing values for the specified index
        between the given start and end dates, both inclusive.

        Parameters
        ----------
        index : str
            Name of the index.

        excel_file : str, optional
            Path to an Excel file to save the DataFrame.

        start_date : str, optional
            Start date in the format 'DD-MMM-YYYY'.
            Defaults to the index's base date if None is provided.

        end_date : str, optional
            End date in the format 'DD-MMM-YYYY'.
            Defaults to the current date if None is provided.

        http_headers : dict, optional
            HTTP headers for the web request. If not provided, defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers`.

        Returns
        -------
        DataFrame
            A DataFrame containing the daily closing values for the index between the specified dates.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # check index name
        if self.is_index_open_source(index):
            index_api = self._index_api.get(index, index)
        else:
            raise Exception(f'"{index}" index data is not available as open-source.')

        # check start date
        if start_date is not None:
            pass
        else:
            start_date = NSEProduct().get_equity_index_base_date(index)
        date_s = Core().string_to_date(start_date)

        # check end date
        if end_date is not None:
            pass
        else:
            end_date = datetime.date.today().strftime('%d-%b-%Y')
        date_e = Core().string_to_date(end_date)

        # check end date is greater than start date
        difference_days = (date_e - date_s).days
        if difference_days >= 0:
            pass
        else:
            raise Exception(f'Start date {start_date} cannot be later than end date {end_date}.')

        # downloaded DataFrame
        df = Core()._download_nse_tri(
            index_api=index_api,
            start_date=start_date,
            end_date=end_date,
            index=index,
            http_headers=http_headers
        )

        # saving the DataFrame
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            df.to_excel(excel_writer, index=False)
            worksheet = excel_writer.sheets['Sheet1']
            worksheet.set_column(0, 1, 12)

        return df

    def update_historical_daily_data(
        self,
        index: str,
        excel_file: str,
        http_headers: typing.Optional[dict[str, str]] = None
    ) -> pandas.DataFrame:

        '''
        Updates historical daily closing values from the last date in the input Excel file
        to the present and saves the aggregated data to the same file.

        Parameters
        ----------
        index : str
            Name of the index.

        excel_file : str
            Path to the Excel file containing existing historical data.

        http_headers : dict, optional
            HTTP headers for the web request. If not provided, defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers`.

        Returns
        -------
        DataFrame
            A DataFrame with updated closing values from the last recorded date to the present.
        '''

        # read the input Excel file
        df = pandas.read_excel(excel_file)
        df['Date'] = df['Date'].apply(
            lambda x: x.date()
        )

        # addition of downloaded DataFrame
        with tempfile.TemporaryDirectory() as tmp_dir:
            add_df = self.download_historical_daily_data(
                index=index,
                excel_file=os.path.join(tmp_dir, 'temporary.xlsx'),
                start_date=df.iloc[-1, 0].strftime('%d-%b-%Y'),
                end_date=datetime.date.today().strftime('%d-%b-%Y'),
                http_headers=http_headers
            )

        # updating the DataFrame
        update_df = pandas.concat([df, add_df]) if isinstance(add_df, pandas.DataFrame) else df
        update_df = update_df.drop_duplicates().reset_index(drop=True)

        # saving the DataFrame
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            update_df.to_excel(excel_writer, index=False)
            worksheet = excel_writer.sheets['Sheet1']
            worksheet.set_column(0, 1, 12)

        return add_df

    def download_daily_summary_equity_closing(
        self,
        excel_file: str,
        http_headers: typing.Optional[dict[str, str]] = None,
        test_mode: bool = False
    ) -> pandas.DataFrame:

        '''
        Returns updated TRI closing values for all NSE indices.

        Parameters
        ----------
        excel_file : str
            Path to an Excel file to save the DataFrame.

        http_headers : dict, optional
            HTTP headers for the web request. Defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers` if not provided.

        test_mode : bool, optional
            If True, the function will use a mocked DataFrame for testing purposes
            instead of the actual data. This parameter is intended for developers
            for testing purposes only and is not recommended for use by end-users.

        Returns
        -------
        DataFrame
            A DataFrame containing updated TRI closing values for all NSE indices.
        '''

        # processing base DataFrame
        base_df = NSEProduct()._dataframe_equity_index
        base_df = base_df.groupby(level='Category').head(2) if test_mode else base_df
        base_df = base_df.reset_index()
        base_df = base_df.drop(columns=['ID', 'API TRI'])
        base_df['Base Date'] = base_df['Base Date'].apply(lambda x: x.date())

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # downloading data
        today = datetime.date.today()
        week_ago = today - datetime.timedelta(days=7)
        end_date = today.strftime('%d-%b-%Y')
        start_date = week_ago.strftime('%d-%b-%Y')
        with tempfile.TemporaryDirectory() as tmp_dir:
            for base_index in base_df.index:
                try:
                    index_df = self.download_historical_daily_data(
                        index=base_df.loc[base_index, 'Index Name'],
                        excel_file=os.path.join(tmp_dir, 'temporary.xlsx'),
                        start_date=start_date,
                        end_date=end_date
                    )
                    base_df.loc[base_index, 'Close Date'] = index_df.iloc[-1, 0]
                    base_df.loc[base_index, 'Close Value'] = index_df.iloc[-1, -1]
                except Exception:
                    print(base_df.loc[base_index, 'Index Name'])
                    base_df.loc[base_index, 'Close Date'] = end_date
                    base_df.loc[base_index, 'Close Value'] = -1000

        # removing error rows from the DataFrame
        base_df = base_df[base_df['Close Value'] != -1000].reset_index(drop=True)

        # saving the DataFrame
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            base_df.to_excel(excel_writer, index=False)
            worksheet = excel_writer.sheets['Sheet1']
            # format columns
            for col_num, df_col in enumerate(base_df.columns):
                if df_col == 'Index Name':
                    worksheet.set_column(col_num, col_num, 60)
                else:
                    worksheet.set_column(col_num, col_num, 15)

        return base_df

    def sort_equity_value_from_launch(
        self,
        input_excel: str,
        output_excel: str,
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by TRI values since launch.

        Parameters
        ----------
        input_excel : str
            Path to the input Excel file.

        output_excel : str
            Path to an Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame sorted in descending order by TRI values since launch.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # sorting DataFrame by TRI values
        df = pandas.read_excel(input_excel)
        df = df.drop(columns=['Category'])
        df = df.sort_values(
            by=['Close Value'],
            ascending=[False]
        )
        df = df.reset_index(drop=True)
        for col_df in df.columns:
            if 'Date' in col_df:
                df[col_df] = df[col_df].apply(lambda x: x.date())
            else:
                pass

        # saving the DataFrame
        with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
            df.to_excel(excel_writer, index=False)
            worksheet = excel_writer.sheets['Sheet1']
            # format columns
            for col_num, col_df in enumerate(df.columns):
                if col_df == 'Index Name':
                    worksheet.set_column(col_num, col_num, 60)
                else:
                    worksheet.set_column(col_num, col_num, 15)

        return df

    def sort_equity_cagr_from_launch(
        self,
        input_excel: str,
        output_excel: str,
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by CAGR (%) since launch.

        Parameters
        ----------
        input_excel : str
            Path to the input Excel file.

        output_excel : str
            Path to an Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame sorted in descending order by CAGR (%) values since launch.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # DataFrame processing
        df = pandas.read_excel(input_excel)
        df = df.drop(columns=['Category'])
        for col_df in df.columns:
            if 'Date' in col_df:
                df[col_df] = df[col_df].apply(lambda x: x.date())
            else:
                pass
        df['Close/Base'] = df['Close Value'] / df['Base Value']
        df['Years'] = list(
            map(
                lambda x, y: dateutil.relativedelta.relativedelta(x, y).years, df['Close Date'], df['Base Date']
            )
        )
        df['Days'] = list(
            map(
                lambda x, y, z: (x - y.replace(year=y.year + z)).days, df['Close Date'], df['Base Date'], df['Years']
            )
        )
        total_years = df['Years'] + (df['Days'] / 365)
        df['CAGR(%)'] = 100 * (pow(df['Close Value'] / df['Base Value'], 1 / total_years) - 1)

        # sorting DataFrame by CAGR (%) values
        df = df.sort_values(
            by=['CAGR(%)', 'Years', 'Days'],
            ascending=[False, False, False]
        )
        df = df.reset_index(drop=True)

        # saving the DataFrame
        with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
            df.to_excel(excel_writer, index=False)
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Sheet1']
            # format columns
            for col_num, col_df in enumerate(df.columns):
                if col_df == 'Index Name':
                    worksheet.set_column(col_num, col_num, 60)
                elif col_df == 'Close Value':
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0'})
                    )
                elif col_df == 'Close/Base':
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0.0'})
                    )
                elif col_df == 'CAGR(%)':
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0.00'})
                    )
                else:
                    worksheet.set_column(col_num, col_num, 15)

        return df

    def category_sort_equity_cagr_from_launch(
        self,
        input_excel: str,
        output_excel: str,
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by CAGR (%) since launch
        within each index category.

        Parameters
        ----------
        input_excel : str
            Path to the input Excel file.

        output_excel : str
            Path to an Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A multi-index DataFrame sorted in descending order by CAGR (%) values since launch
            within each index category.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # DataFrame processing
        df = pandas.read_excel(input_excel)
        for col_df in df.columns:
            if 'Date' in col_df:
                df[col_df] = df[col_df].apply(lambda x: x.date())
            else:
                pass
        df['Close/Base'] = df['Close Value'] / df['Base Value']
        df['Years'] = list(
            map(
                lambda x, y: dateutil.relativedelta.relativedelta(x, y).years, df['Close Date'], df['Base Date']
            )
        )
        df['Days'] = list(
            map(
                lambda x, y, z: (x - y.replace(year=y.year + z)).days, df['Close Date'], df['Base Date'], df['Years']
            )
        )
        total_years = df['Years'] + (df['Days'] / 365)
        df['CAGR(%)'] = 100 * (pow(df['Close Value'] / df['Base Value'], 1 / total_years) - 1)

        # Convert 'Category' column to categorical data types with a defined order
        categories = list(df['Category'].unique())
        df['Category'] = pandas.Categorical(
            df['Category'],
            categories=categories,
            ordered=True
        )

        # Sorting Dataframe
        df = df.sort_values(
            by=['Category', 'CAGR(%)', 'Years', 'Days'],
            ascending=[True, False, False, False]
        )
        dataframes = []
        for category in categories:
            category_df = df[df['Category'] == category]
            category_df = category_df.drop(columns=['Category']).reset_index(drop=True)
            dataframes.append(category_df)
        output = pandas.concat(
            dataframes,
            keys=[word.upper() for word in categories],
            names=['Category', 'ID']
        )

        # saving the DataFrame
        with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
            output.to_excel(excel_writer, index=True)
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Sheet1']
            # number of columns for DataFrame indices
            index_cols = len(output.index.names)
            # format columns
            worksheet.set_column(0, index_cols - 1, 15)
            for col_num, col_df in enumerate(output.columns):
                if col_df == 'Index Name':
                    worksheet.set_column(index_cols + col_num, index_cols + col_num, 60)
                elif col_df == 'Close Value':
                    worksheet.set_column(
                        index_cols + col_num, index_cols + col_num, 15,
                        workbook.add_format({'num_format': '#,##0'})
                    )
                elif col_df == 'Close/Base':
                    worksheet.set_column(
                        index_cols + col_num, index_cols + col_num, 15,
                        workbook.add_format({'num_format': '#,##0.0'})
                    )
                elif col_df == 'CAGR(%)':
                    worksheet.set_column(
                        index_cols + col_num, index_cols + col_num, 15,
                        workbook.add_format({'num_format': '#,##0.00'})
                    )
                else:
                    worksheet.set_column(index_cols + col_num, index_cols + col_num, 15)
            # Dataframe colors
            get_colormap = matplotlib.colormaps.get_cmap('Pastel2')
            colors = [
                get_colormap(count / len(dataframes)) for count in range(len(dataframes))
            ]
            hex_colors = [
                '{:02X}{:02X}{:02X}'.format(*[int(num * 255) for num in color]) for color in colors
            ]
            # coloring of DataFrames
            start_col = index_cols - 1
            end_col = index_cols + len(output.columns) - 1
            start_row = 1
            for df, color in zip(dataframes, hex_colors):
                color_format = workbook.add_format({'bg_color': color})
                end_row = start_row + len(df) - 1
                worksheet.conditional_format(
                    start_row, start_col, end_row, end_col,
                    {'type': 'no_blanks', 'format': color_format}
                )
                start_row = end_row + 1

        return output

    def compare_cagr_over_price_from_launch(
        self,
        tri_excel: str,
        price_excel: str,
        output_excel: str
    ) -> pandas.DataFrame:

        '''
        Compares the CAGR (%) between TRI and Price for NSE indices since launch.

        Parameters
        ----------
        tri_excel : str
            Path to the Excel file obtained from :meth:`BharatFinTrack.NSETRI.sort_equity_cagr_from_launch` method.

        price_excel : str
            Path to the Excel file obtained from :meth:`BharatFinTrack.NSEIndex.sort_equity_cagr_from_launch` method.

        output_excel : str
            Path to an Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the difference in CAGR (%) between TRI to Price since launch.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # TRI Dataframe
        tri_df = pandas.read_excel(tri_excel)
        tri_df = tri_df.rename(
            columns={'CAGR(%)': 'TRI-CAGR(%)'}
        )
        for date_col in ['Base Date', 'Close Date']:
            tri_df[date_col] = tri_df[date_col].apply(lambda x: x.date())

        # Price DataFrame
        price_df = pandas.read_excel(price_excel)
        price_df = price_df.rename(
            columns={'CAGR(%)': 'PRICE-CAGR(%)'}
        )

        # merge TRI and PRICE DataFrames
        df = tri_df.merge(
            right=price_df[['Index Name', 'PRICE-CAGR(%)']],
            on='Index Name',
            how='left'
        )
        df['Difference(%)'] = tri_df['TRI-CAGR(%)'] - price_df['PRICE-CAGR(%)']
        df = df.drop(
            columns=['Close Value', 'Close/Base']
        )

        # saving the DataFrame
        with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
            df.to_excel(excel_writer, index=False)
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Sheet1']
            # format columns
            for col_num, col_df in enumerate(df.columns):
                if col_df == 'Index Name':
                    worksheet.set_column(col_num, col_num, 60)
                elif col_df.endswith('(%)'):
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0.00'})
                    )
                else:
                    worksheet.set_column(col_num, col_num, 15)

        return df

    def yearwise_sip_analysis(
        self,
        input_excel: str,
        monthly_invest: int,
        output_excel: str
    ) -> pandas.DataFrame:

        '''
        Calculates the year-wise closing value, growth multiples, and annualized XIRR (%)
        of a fixed monthly SIP, based on contributions made on the first date of each month.

        Parameters
        ----------
        input_excel : str
            Path to the Excel file obtained from :meth:`BharatFinTrack.NSETRI.download_historical_daily_data`
            and :meth:`BharatFinTrack.NSETRI.update_historical_daily_data` methods.

        monthly_invest : int
            Fixed investment amount contributed on the first date of each month.

        output_excel : str
            Path to an Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the year-wise closing value, growth multiples,
            and annualized XIRR (%) for the fixed SIP investment.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # input DataFrame
        df = pandas.read_excel(input_excel)
        df['Date'] = df['Date'].apply(lambda x: x.date())

        # start and end dates
        start_date = df['Date'].min()
        end_date = df['Date'].max()

        # monthly first date
        month_1d = pandas.date_range(
            start=start_date,
            end=end_date,
            freq='MS'
        )
        month_1d = list(map(lambda x: x.date(), month_1d))
        month_1d = pandas.Series([start_date] + month_1d + [end_date]).unique()

        # DataFrame of monthly open and close value
        month_df = pandas.DataFrame(
            columns=['Date', 'Open', 'Close']
        )
        for idx, dates in enumerate(zip(month_1d[:-1], month_1d[1:])):
            idx_df = df[(df['Date'] >= dates[0]) & (df['Date'] < dates[1])]
            month_df.loc[idx, 'Date'] = idx_df.iloc[0, 0]
            month_df.loc[idx, 'Open'] = idx_df.iloc[0, 1]
            month_df.loc[idx, 'Close'] = idx_df.iloc[-1, -1]

        # date difference
        date_diff = dateutil.relativedelta.relativedelta(end_date, start_date)
        year_diff = date_diff.years

        # SIP DataFrame
        index_divisor = 1000
        sip_df = pandas.DataFrame()
        for idx in range(year_diff + 1):
            # year-wise SIP investment
            if idx < year_diff:
                sip_year = idx + 1
                sip_start = end_date.replace(year=end_date.year - sip_year)
                yi_df = month_df[(month_df['Date'] >= sip_start) & (month_df['Date'] < end_date)].reset_index(drop=True)
            else:
                sip_year = round(year_diff + (end_date.replace(year=end_date.year - year_diff) - start_date).days / 365, 1)
                yi_df = month_df.copy()
            yi_df['Invest'] = monthly_invest
            yi_df['Cumm-Invest'] = yi_df['Invest'].cumsum()
            open_nav = yi_df['Open'] / index_divisor
            yi_df['Unit'] = (yi_df['Invest'] / open_nav)
            yi_df['Cum-Unit'] = yi_df['Unit'].cumsum()
            close_nav = yi_df['Close'] / index_divisor
            yi_df['Value'] = (yi_df['Cum-Unit'] * close_nav)
            # year-wise SIP summary
            sip_df.loc[idx, 'Year'] = sip_year
            sip_df.loc[idx, 'Start Date'] = yi_df.iloc[0, 0]
            sip_df.loc[idx, 'Invest'] = yi_df.iloc[-1, -4]
            sip_df.loc[idx, 'Close Date'] = end_date
            sip_df.loc[idx, 'Value'] = yi_df.iloc[-1, -1]
            sip_df.loc[idx, 'Multiple (X)'] = sip_df.loc[idx, 'Value'] / sip_df.loc[idx, 'Invest']
            sip_dates = list(yi_df['Date']) + [end_date]
            sip_transactions = list(-1 * yi_df['Invest']) + [yi_df['Value'].iloc[-1]]
            xirr = pyxirr.xirr(zip(sip_dates, sip_transactions))
            sip_df.loc[idx, 'XIRR (%)'] = 100 * (xirr if xirr is not None else 0.0)

        # drop duplicates row if any
        sip_df = sip_df.drop_duplicates(ignore_index=True)

        # saving DataFrame
        with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
            sip_df.to_excel(excel_writer, index=False)
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Sheet1']
            # format columns
            for col_num, col_df in enumerate(sip_df.columns):
                if any(col_df.endswith(i) for i in ['(%)', 'Year', '(X)']):
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0.0'})
                    )
                else:
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0'})
                    )

        return sip_df

    def sip_summary_from_given_date(
        self,
        excel_file: str,
        start_year: int,
        start_month: int,
        monthly_invest: int
    ) -> dict[str, str]:

        '''
        Calculates the closing value, growth multiples, and annualized XIRR (%) of a fixed monthly
        SIP starting from a specified date, based on contributions made on the first date of each month.

        Parameters
        ----------
        excel_file : str
            Path to the Excel file obtained from :meth:`BharatFinTrack.NSETRI.download_historical_daily_data`
            and :meth:`BharatFinTrack.NSETRI.update_historical_daily_data` methods.

        start_year : int
            Year when the SIP begins.

        start_month : int
            Month (1 to 12) when the SIP begins.

        monthly_invest : int
            Fixed investment amount contributed on the first date of each month

        Returns
        -------
        dict
            A dictionary containing the closing value, growth multiples, and annualized XIRR (%)
            for the fixed SIP investment starting from a specified date.
        '''

        # SIP summary
        summary = {}

        # input DataFrame
        df = pandas.read_excel(excel_file)
        df['Date'] = df['Date'].apply(lambda x: x.date())

        # start and end dates
        start_date = datetime.date(
            year=start_year,
            month=start_month,
            day=1
        )
        end_date = df.iloc[-1, 0]

        # filtered DataFrame
        if start_date > df.iloc[-1, 0]:
            raise Exception(
                'Given year and month return an empty DataFrame.'
            )
        elif df.iloc[0, 0] <= start_date <= df.iloc[-1, 0]:
            summary['Start date'] = start_date.strftime('%d-%b-%Y')
        else:
            summary['Given date'] = start_date.strftime('%d-%b-%Y')
            df = df[(df['Date'] >= start_date)].reset_index(drop=True)
            start_date = df.iloc[0, 0]
            summary['Actual start date'] = start_date.strftime('%d-%b-%Y')

        # month first date
        month_1d = pandas.date_range(
            start=start_date,
            end=end_date,
            freq='MS'
        )
        month_1d = list(map(lambda x: x.date(), month_1d))
        month_1d = pandas.Series([start_date] + month_1d + [end_date]).unique()

        # DataFrame of monthly open and close value
        month_df = pandas.DataFrame(
            columns=['Date', 'Open', 'Close']
        )
        for idx, dates in enumerate(zip(month_1d[:-1], month_1d[1:])):
            idx_df = df[(df['Date'] >= dates[0]) & (df['Date'] < dates[1])]
            month_df.loc[idx, 'Date'] = idx_df.iloc[0, 0]
            month_df.loc[idx, 'Open'] = idx_df.iloc[0, 1]
            month_df.loc[idx, 'Close'] = idx_df.iloc[-1, -1]

        # SIP parameters
        index_divisor = 1000
        date_diff = dateutil.relativedelta.relativedelta(end_date, start_date)
        month_df['Invest'] = monthly_invest
        month_df['Cumm-Invest'] = month_df['Invest'].cumsum()
        open_nav = month_df['Open'] / index_divisor
        month_df['Unit'] = (month_df['Invest'] / open_nav)
        month_df['Cum-Unit'] = month_df['Unit'].cumsum()
        close_nav = month_df['Close'] / index_divisor
        month_df['Value'] = (month_df['Cum-Unit'] * close_nav)
        # year-wise SIP summary
        summary['Duration'] = f'{date_diff.years} years, {date_diff.months} months, {date_diff.days} days'
        summary['Invest'] = f'{month_df.iloc[-1, -4]:.0f}'
        summary['Value'] = f'{month_df.iloc[-1, -1]:.0f}'
        summary['Multiple (X)'] = f'{month_df.iloc[-1, -1] / month_df.iloc[-1, -4]:.1f}'
        sip_dates = list(month_df['Date']) + [end_date]
        sip_transactions = list(-1 * month_df['Invest']) + [month_df['Value'].iloc[-1]]
        xirr = pyxirr.xirr(zip(sip_dates, sip_transactions))
        xirr_p = 100 * (xirr if xirr is not None else 0.0)
        summary['XIRR (%)'] = f'{xirr_p:.1f}'

        return summary

    def sip_growth_comparison_across_indices(
        self,
        indices: list[str],
        folder_path: str,
        excel_file: str,
    ) -> pandas.DataFrame:

        '''
        Generates a DataFrame that compares SIP investment growth on the
        first date of each month across multiple indices over the years.
        The output DataFrame is saved to an Excel file, where the cells with
        the highest growth among indices for each year are highlighted in green-yellow,
        and those with the lowest growth are highlighted in sandy brown.

        Additionally, a scoring mechanism is implemented for the indices based on their growth values.
        For each year, indices are ranked in ascending order of growth, with the lowest value
        receiving the lowest score (1), and the highest value receiving the highest score.
        The total scores for each index are calculated by summing their yearly scores.
        Indices are then sorted in descending order based on their total scores,
        and the results are converted into a DataFrame with columns 'Index Name' and 'Score'.

        Parameters
        ----------
        indices : list
            A list of index names to compare in the SIP growth.

        folder_path : str
            Path to the directory containing Excel files with historical data for each index. Each Excel file must be
            named as '{index}.xlsx' corresponding to the index names provided in the `indices` list. These files should
            be obtained from :meth:`BharatFinTrack.NSETRI.download_historical_daily_data` or
            :meth:`BharatFinTrack.NSETRI.update_historical_daily_data`.

        excel_file : str
            Path to an Excel file to save the output DataFrames.

        Returns
        -------
        DataFrame
            A DataFrame containing the index names and their total scores.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # monthly investment amount
        monthly_invest = 1000

        # SIP dataframe of index
        dataframes = []
        with tempfile.TemporaryDirectory() as tmp_dir:
            for index in indices:
                index_excel = os.path.join(folder_path, f'{index}.xlsx')
                df = NSETRI().yearwise_sip_analysis(
                    input_excel=index_excel,
                    monthly_invest=monthly_invest,
                    output_excel=os.path.join(tmp_dir, 'output.xlsx')
                )
                dataframes.append(df)

        # check equal close date for all DataFrames
        close_date = dataframes[0]['Close Date'].iloc[0]
        equal_closedate = all(map(lambda df: df['Close Date'].iloc[0] == close_date, dataframes))
        if equal_closedate:
            pass
        else:
            raise Exception('Last date must be equal across all indices in the Excel files.')

        # filtered dataframes
        common_year = min(
            map(lambda df: int(df['Year'].max()), dataframes)
        )
        dataframes = [
            df[df['Year'] <= common_year] for df in dataframes
        ]
        dataframes = [
            df.drop(columns=['Invest', 'Value', 'XIRR (%)']) for df in dataframes
        ]
        dataframes = [
            df.rename(columns={'Multiple (X)': f'{index} (X)'}) for df, index in zip(dataframes, indices)
        ]

        # mergeing the DataFrames
        merged_df = dataframes[0]
        common_cols = list(merged_df.columns)[:-1]
        for df in dataframes[1:]:
            merged_df = pandas.merge(
                left=merged_df,
                right=df,
                on=common_cols,
                how='inner'
            )

        # assing score to indices growth returns
        score_df = merged_df.copy()
        score_df = score_df.iloc[:, len(common_cols):]
        for idx, row in score_df.iterrows():
            sort_growth = row.sort_values(ascending=True).index
            score_indices = range(1, len(sort_growth) + 1)
            score_df.loc[idx, sort_growth] = score_indices

        # aggregate DataFrame of sorted total score
        aggregate_df = score_df.sum().sort_values(ascending=False).reset_index()
        aggregate_df.columns = ['Index Name', 'Score']
        aggregate_df['Index Name'] = aggregate_df['Index Name'].apply(lambda x: x.replace(' (X)', ''))

        # rounding of column values to catch exact maximum and minimum with floating point precision
        for col in merged_df.columns:
            if col.endswith('(X)'):
                merged_df[col] = merged_df[col].round(5)
            else:
                pass

        # saving DataFrames
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            ##################
            # merged DataFrame
            merged_df.to_excel(
                excel_writer=excel_writer,
                index=False,
                sheet_name='Multiple(X)'
            )
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Multiple(X)']
            worksheet.set_column(
                0, len(common_cols) - 1, 15
            )
            worksheet.set_column(
                len(common_cols), merged_df.shape[1] - 1, 15,
                workbook.add_format({'num_format': '#,##0.0'})
            )
            # header formatting
            header_format = workbook.add_format(
                {
                    'bold': True,
                    'text_wrap': True,
                    'align': 'center',
                    'valign': 'vcenter'
                }
            )
            for col_num, col_df in enumerate(merged_df.columns):
                worksheet.write(0, col_num, col_df, header_format)
            # formatting for maximum and minimum value in each row
            for row in range(merged_df.shape[0]):
                # minimum value
                worksheet.conditional_format(
                    row + 1, len(common_cols), row + 1, merged_df.shape[1] - 1,
                    {
                        'type': 'cell',
                        'criteria': 'equal to',
                        'value': merged_df.iloc[row, len(common_cols):].min(),
                        'format': workbook.add_format({'bg_color': '#F4A460'})
                    }
                )
                # maximim value
                worksheet.conditional_format(
                    row + 1, len(common_cols), row + 1, merged_df.shape[1] - 1,
                    {
                        'type': 'cell',
                        'criteria': 'equal to',
                        'value': merged_df.iloc[row, len(common_cols):].max(),
                        'format': workbook.add_format({'bg_color': '#ADFF2F'})
                    }
                )
            ##################
            # score DataFrame
            aggregate_df.to_excel(
                excel_writer=excel_writer,
                index=False,
                sheet_name='Score'
            )
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Score']
            worksheet.set_column(0, 0, 75)
            worksheet.set_column(1, 1, 15)

        return aggregate_df

    def sip_xirr_comparison_across_indices(
        self,
        indices: list[str],
        folder_path: str,
        excel_file: str,
    ) -> pandas.DataFrame:

        '''
        Generates a DataFrame that compares XIRR (%) of SIP growth on the
        first date of each month across multiple indices over the years.
        The output DataFrame is saved to an Excel file, where the cells with
        the highest XIRR (%) among indices for each year are highlighted in green-yellow,
        and those with the lowest XIRR (%) are highlighted in sandy brown.

        Additionally, a scoring mechanism is implemented for the indices based on their XIRR (%) values.
        For each year, indices are ranked in ascending order of XIRR (%), with the lowest value
        receiving the lowest score (1), and the highest value receiving the highest score.
        The total scores for each index are calculated by summing their yearly scores.
        Indices are then sorted in descending order based on their total scores,
        and the results are converted into a DataFrame with columns 'Index Name' and 'Score'.

        Parameters
        ----------
        indices : list
            A list of index names to compare in the SIP XIRR (%).

        folder_path : str
            Path to the directory containing Excel files with historical data for each index. Each Excel file must be
            named as '{index}.xlsx' corresponding to the index names provided in the `indices` list. These files should
            be obtained from :meth:`BharatFinTrack.NSETRI.download_historical_daily_data` or
            :meth:`BharatFinTrack.NSETRI.update_historical_daily_data`.

        excel_file : str
            Path to an Excel file to save the output DataFrames.

        Returns
        -------
        DataFrame
            A DataFrame containing the index names and their total scores.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # monthly investment amount
        monthly_invest = 1000

        # SIP dataframe of index
        dataframes = []
        with tempfile.TemporaryDirectory() as tmp_dir:
            for index in indices:
                index_excel = os.path.join(folder_path, f'{index}.xlsx')
                df = NSETRI().yearwise_sip_analysis(
                    input_excel=index_excel,
                    monthly_invest=monthly_invest,
                    output_excel=os.path.join(tmp_dir, 'output.xlsx')
                )
                dataframes.append(df)

        # check equal close date for all DataFrames
        close_date = dataframes[0]['Close Date'].iloc[0]
        equal_closedate = all(map(lambda df: df['Close Date'].iloc[0] == close_date, dataframes))
        if equal_closedate:
            pass
        else:
            raise Exception('Last date must be equal across all indices in the Excel files.')

        # filtered dataframes
        common_year = min(
            map(lambda df: int(df['Year'].max()), dataframes)
        )
        dataframes = [
            df[df['Year'] <= common_year] for df in dataframes
        ]
        dataframes = [
            df.drop(columns=['Invest', 'Value', 'Multiple (X)']) for df in dataframes
        ]
        dataframes = [
            df.rename(columns={'XIRR (%)': f'{index} (XIRR)'}) for df, index in zip(dataframes, indices)
        ]

        # mergeing the DataFrames
        merged_df = dataframes[0]
        common_cols = list(merged_df.columns)[:-1]
        for df in dataframes[1:]:
            merged_df = pandas.merge(
                left=merged_df,
                right=df,
                on=common_cols,
                how='inner'
            )

        # assing score to indices growth returns
        score_df = merged_df.copy()
        score_df = score_df.iloc[:, len(common_cols):]
        for idx, row in score_df.iterrows():
            sort_growth = row.sort_values(ascending=True).index
            score_indices = range(1, len(sort_growth) + 1)
            score_df.loc[idx, sort_growth] = score_indices

        # aggregate DataFrame of sorted total score
        aggregate_df = score_df.sum().sort_values(ascending=False).reset_index()
        aggregate_df.columns = ['Index Name', 'Score']
        aggregate_df['Index Name'] = aggregate_df['Index Name'].apply(lambda x: x.replace(' (XIRR)', ''))

        # rounding of column values to catch exact maximum and minimum with floating point precision
        for col in merged_df.columns:
            if col.endswith('(XIRR)'):
                merged_df[col] = merged_df[col].round(5)
            else:
                pass

        # saving DataFrames
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            ##################
            # merged DataFrame
            merged_df.to_excel(
                excel_writer=excel_writer,
                index=False,
                sheet_name='XIRR(%)'
            )
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['XIRR(%)']
            worksheet.set_column(
                0, len(common_cols) - 1, 15
            )
            worksheet.set_column(
                len(common_cols), merged_df.shape[1] - 1, 15,
                workbook.add_format({'num_format': '#,##0.0'})
            )
            # header formatting
            header_format = workbook.add_format(
                {
                    'bold': True,
                    'text_wrap': True,
                    'align': 'center',
                    'valign': 'vcenter'
                }
            )
            for col_num, col_df in enumerate(merged_df.columns):
                worksheet.write(0, col_num, col_df, header_format)
            # formatting for maximum and minimum value in each row
            for row in range(merged_df.shape[0]):
                # minimum value
                worksheet.conditional_format(
                    row + 1, len(common_cols), row + 1, merged_df.shape[1] - 1,
                    {
                        'type': 'cell',
                        'criteria': 'equal to',
                        'value': merged_df.iloc[row, len(common_cols):].min(),
                        'format': workbook.add_format({'bg_color': '#F4A460'})
                    }
                )
                # maximim value
                worksheet.conditional_format(
                    row + 1, len(common_cols), row + 1, merged_df.shape[1] - 1,
                    {
                        'type': 'cell',
                        'criteria': 'equal to',
                        'value': merged_df.iloc[row, len(common_cols):].max(),
                        'format': workbook.add_format({'bg_color': '#ADFF2F'})
                    }
                )
            ##################
            # score DataFrame
            aggregate_df.to_excel(
                excel_writer=excel_writer,
                index=False,
                sheet_name='Score'
            )
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Score']
            worksheet.set_column(0, 0, 75)
            worksheet.set_column(1, 1, 15)

        return aggregate_df



import os
import datetime
import pandas
from .core import Core


class NSEProduct:

    '''
    Provides functionality for accessing the characteristics of
    NSE related financial products.
    '''

    @property
    def _dataframe_equity_index(
        self
    ) -> pandas.DataFrame:

        '''
        Returns a multi-index DataFrame containing
        the characteristics of equity indices.
        '''

        file_path = os.path.join(
            os.path.dirname(__file__), 'data', 'equity_indices.xlsx'
        )

        dataframes = pandas.read_excel(
            io=file_path,
            sheet_name=None
        )

        df = pandas.concat(
            dataframes,
            names=['Category', 'ID']
        )

        return df

    def save_dataframe_equity_index_parameters(
        self,
        excel_file: str
    ) -> pandas.DataFrame:

        '''
        Saves a multi-index DataFrame containing the characteristics
        of equity indices to an Excel file.

        Parameters
        ----------
        excel_file : str
            Path of an Excel file to save the multi-index DataFrame.

        Returns
        -------
        DataFrame
            The multi-index DataFrame containing the characteristics of equity indices.
        '''

        # Excel file extension
        excel_ext = Core()._excel_file_extension(
            file_path=excel_file
        )

        # saving the multi-index dataframe
        if excel_ext == '.xlsx':
            df = self._dataframe_equity_index
            df = df[df.columns[:3]]
            df['Base Date'] = df['Base Date'].apply(lambda x: x.date())
            with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
                df.to_excel(excel_writer, index=True)
                workbook = excel_writer.book
                worksheet = excel_writer.sheets['Sheet1']
                worksheet.set_column(0, 1, 12, workbook.add_format({'align': 'center'}))
                worksheet.set_column(2, 2, 60, workbook.add_format({'align': 'left'}))
                worksheet.set_column(3, 4, 12, workbook.add_format({'align': 'right'}))
        else:
            raise Exception(
                f'Input file extension "{excel_ext}" does not match the required ".xlsx".'
            )

        return df

    @property
    def equity_index_category(
        self
    ) -> list[str]:

        '''
        Returns a list of categories for NSE equity indices.
        '''

        df = self._dataframe_equity_index.reset_index()
        output = list(df['Category'].unique())

        return output

    def get_equity_indices_by_category(
        self,
        category: str
    ) -> list[str]:

        '''
        Returns a list of NSE equity indices for a specified index category.

        Parameters
        ----------
        category : str
            The category of NSE indices.

        Returns
        -------
        list
            A list containing the equity indices for the specified category.
        '''

        if category in self.equity_index_category:
            df = self._dataframe_equity_index.reset_index()
            df = df[df['Category'] == category]
            output = list(df['Index Name'].sort_values())
        else:
            raise Exception(f'Input category "{category}" does not exist.')

        return output

    @property
    def all_equity_indices(
        self
    ) -> list[str]:

        '''
        Returns a list of equity indices for all categories.
        '''

        df = self._dataframe_equity_index
        output = list(df['Index Name'].sort_values())

        return output

    def is_index_exist(
        self,
        index: str
    ) -> bool:

        '''
        Returns whether the index exists in the list of equity indices.

        Parameters
        ----------
        index : str
            Name of the index.

        Returns
        -------
        bool
            True if the index exists, False otherwise.
        '''

        output = index in self.all_equity_indices

        return output

    def get_equity_index_base_date(
        self,
        index: str
    ) -> str:

        '''
        Returns the base date for a specified NSE equity index.

        Parameters
        ----------
        index : str
            Name of the index.

        Returns
        -------
        str
            The base date of the index in 'DD-MMM-YYYY' format.
        '''

        df = self._dataframe_equity_index
        df = df[['Index Name', 'Base Date']]
        df = df[df['Index Name'] == index]

        if df.shape[0] > 0 and isinstance(df.iloc[-1, -1], datetime.datetime):
            output = df.iloc[-1, -1].strftime('%d-%b-%Y')
        else:
            raise Exception(f'"{index}" index does not exist.')

        return output

    def get_equity_index_base_value(
        self,
        index: str
    ) -> float:

        '''
        Returns the base value for a specified NSE equity index.

        Parameters
        ----------
        index : str
            Name of the index.

        Returns
        -------
        float
            The base value of the index.
        '''

        df = self._dataframe_equity_index
        df = df[['Index Name', 'Base Value']]
        df = df[df['Index Name'] == index]

        if df.shape[0] > 0:
            output = float(df.iloc[-1, -1])
        else:
            raise Exception(f'"{index}" index does not exist.')

        return output



import os
import tempfile
import pandas
import matplotlib
import matplotlib.pyplot
import matplotlib.figure
from .core import Core
from .nse_tri import NSETRI


class Visual:

    '''
    Provides functionality for plotting data.
    '''

    def _mi_df_bar_closing_with_category(
        self,
        df: pandas.DataFrame,
        close_type: str,
        index_type: str,
        figure_title: str,
        figure_file: str
    ) -> matplotlib.figure.Figure:

        '''
        Helper function to create a bar plot of the closing values (PRICE/TRI)
        of NSE indices from a multi-index DataFrame.
        '''

        # check validity of input figure file path
        check_file = Core().is_valid_figure_extension(figure_file)
        if check_file:
            pass
        else:
            raise Exception('Input figure file extension is not supported.')

        # check close value type
        if close_type in ['PRICE', 'TRI']:
            pass
        else:
            raise Exception(f'Invalid indices close value type: {close_type}; must be one of [PRICE, TRI].')

        # catgory of indices
        categories = df.index.get_level_values('Category').unique()

        # close date
        close_date = df['Close Date'].iloc[0].strftime('%d-%b-%Y')

        # color for NSE indices category
        colormap = matplotlib.colormaps.get_cmap('Set2')
        category_color = {
            categories[count]: colormap(count / len(categories)) for count in range(len(categories))
        }

        # figure
        fig_height = int(len(df) / 3.5) + 1 if len(df) >= 18 else 5
        xtick_gap = 10000
        xaxis_max = int(((df['Close Value'].max() + 20000) / xtick_gap) + 1) * xtick_gap
        fig_width = int((xaxis_max / xtick_gap) * 1.5)
        figure = matplotlib.pyplot.figure(
            figsize=(fig_width, fig_height)
        )
        subplot = figure.subplots(1, 1)

        # plotting indices closing values
        categories_legend = set()
        for count, (index, row) in enumerate(df.iterrows()):
            category = index[0]
            color = category_color[category]
            if category not in categories_legend:
                subplot.barh(
                    row['Index Name'], row['Close Value'],
                    color=color,
                    label=category
                )
                categories_legend.add(category)
            else:
                subplot.barh(
                    row['Index Name'], row['Close Value'],
                    color=color
                )
            age = row['Years'] + (row['Days'] / 365)
            bar_label = f"({row['CAGR(%)']:.1f}%,{round(row['Close/Base'])}X,{age:.1f}Y)"
            subplot.text(
                row['Close Value'] + 100, count, bar_label,
                va='center',
                fontsize=10
            )

        # x-axis customization
        subplot.set_xlim(0, xaxis_max)
        xticks = range(0, xaxis_max + 1, xtick_gap)
        subplot.set_xticks(
            ticks=xticks
        )
        xticklabels = [
            f'{int(val / 1000)}k' for val in xticks
        ]
        subplot.set_xticklabels(
            labels=xticklabels,
            fontsize=12
        )
        subplot.tick_params(
            axis='x', which='both',
            direction='in', length=6, width=1,
            top=True, bottom=True,
            labeltop=True, labelbottom=True
        )
        subplot.grid(
            visible=True,
            which='major', axis='x',
            color='gray',
            linestyle='--', linewidth=0.3
        )
        subplot.set_xlabel(
            f'Close Value (Date: {close_date})',
            fontsize=15,
            labelpad=15
        )

        # reverse y-axis
        subplot.invert_yaxis()

        # y-axis customization
        subplot.set_ylabel(
            f'{index_type} Index',
            fontsize=15,
            labelpad=15
        )
        subplot.set_ylim(len(df), -1)

        # legend
        subplot.legend(
            title='Index Category',
            loc='lower right',
            fontsize=12,
            title_fontsize=12
        )

        # figure customization
        figure.suptitle(
            figure_title,
            fontsize=15,
            y=1
        )
        figure.tight_layout()
        figure.savefig(
            figure_file,
            bbox_inches='tight'
        )

        # close the figure to prevent duplicate plots from displaying
        matplotlib.pyplot.close(figure)

        return figure

    def plot_cagr_filtered_indices_by_category(
        self,
        excel_file: str,
        close_type: str,
        figure_file: str,
        threshold_cagr: float = 10.0,
        index_type: str = 'NSE Equity'
    ) -> matplotlib.figure.Figure:

        '''
        Returns a bar plot of the closing values (PRICE/TRI) of NSE indices grouped by categoty,
        filtered by a specified threshold CAGR (%) since their launch.

        Parameters
        ----------
        excel_file : str
            Path of the input Excel file containing the data.

        close_type : str
            Type of closing value for indices, either 'PRICE' or 'TRI'.

        figure_file : str
            File Path to save the output figure.

        threshold_cagr : float, optional
            Only plot indices with a CAGR (%) higher than the specified threshold.
            Default is 10.

        index_type : str, optional
            Type of index. Default is 'NSE Equity'.

        Returns
        -------
        Figure
            A bar plot displaying closing values of NSE indices, along with
            CAGR (%), Multiplier (X), and Age (Y) since launch.
        '''

        # input DataFrame
        df = pandas.read_excel(excel_file, index_col=[0, 1])
        df = df[df['CAGR(%)'] >= threshold_cagr]

        # check filtered dataframe
        if len(df) == 0:
            raise Exception('Threshold values return an empty DataFrame.')
        else:
            pass

        # figure
        figure_title = (
            f'Category-wise Threshold CAGR (>= {threshold_cagr} %) Since Launch: '
            f'Closing {close_type} (Bar) with CAGR (%), Multiplier (X), and Age (Y)'
        )
        figure = self._mi_df_bar_closing_with_category(
            df=df,
            close_type=close_type,
            index_type=index_type,
            figure_title=figure_title,
            figure_file=figure_file
        )

        return figure

    def plot_top_cagr_indices_by_category(
        self,
        excel_file: str,
        close_type: str,
        figure_file: str,
        top_cagr: int = 5,
        index_type: str = 'NSE Equity'
    ) -> matplotlib.figure.Figure:

        '''
        Returns a bar plot of the closing values (PRICE/TRI) of NSE indices
        with the top CAGR (%) since their launch, grouped by category.

        Parameters
        ----------
        excel_file : str
            Path of the input Excel file containing the data.

        close_type : str
            Type of closing value for indices, either 'PRICE' or 'TRI'.

        figure_file : str
            File Path to save the output figure.

        top_cagr : int, optional
            The number of top indices by CAGR (%) to plot for each category.
            Default is 5.

        index_type : str, optional
            Type of index. Default is 'NSE Equity'.

        Returns
        -------
        Figure
            A bar plot displaying closing values of NSE indices, along with
            CAGR (%), Multiplier (X), and Age (Y) since launch.
        '''

        # input DataFrame
        df = pandas.read_excel(excel_file, index_col=[0, 1])
        df = df.groupby(level='Category').head(top_cagr)

        # figure
        figure_title = (
            f'Category-wise Top {top_cagr} CAGR (%) Since Launch: '
            f'Closing {close_type} (Bar) with CAGR (%), Multiplier (X), and Age (Y)'
        )
        figure = self._mi_df_bar_closing_with_category(
            df=df,
            close_type=close_type,
            index_type=index_type,
            figure_title=figure_title,
            figure_file=figure_file
        )

        return figure

    def _df_bar_closing(
        self,
        df: pandas.DataFrame,
        close_type: str,
        index_type: str,
        figure_title: str,
        figure_file: str
    ) -> matplotlib.figure.Figure:

        '''
        Helper function to create a bar plot of the closing values (PRICE/TRI)
        for NSE indices from a DataFrame where index categories are not specified.
        '''

        # check validity of input figure file path
        check_file = Core().is_valid_figure_extension(figure_file)
        if check_file:
            pass
        else:
            raise Exception('Input figure file extension is not supported.')

        # check close value type
        if close_type in ['PRICE', 'TRI']:
            pass
        else:
            raise Exception(f'Invalid indices close value type: {close_type}; must be one of [PRICE, TRI].')

        # close date
        close_date = df['Close Date'].iloc[0].strftime('%d-%b-%Y')

        # figure
        fig_height = int(len(df) / 3.5) + 1 if len(df) >= 18 else 5
        xtick_gap = 10000
        xaxis_max = int(((df['Close Value'].max() + 20000) / xtick_gap) + 1) * xtick_gap
        fig_width = int((xaxis_max / xtick_gap) * 1.5)
        figure = matplotlib.pyplot.figure(
            figsize=(fig_width, fig_height)
        )
        subplot = figure.subplots(1, 1)

        # plotting indices closing values
        bar_color = 'lawngreen' if close_type == 'TRI' else 'cyan'
        for count, (index, row) in enumerate(df.iterrows()):
            subplot.barh(
                row['Index Name'], row['Close Value'],
                color=bar_color
            )
            age = row['Years'] + (row['Days'] / 365)
            bar_label = f"({row['CAGR(%)']:.1f}%,{round(row['Close/Base'])}X,{age:.1f}Y)"
            subplot.text(
                row['Close Value'] + 100, count, bar_label,
                va='center',
                fontsize=10
            )

        # x-axis customization
        subplot.set_xlim(0, xaxis_max)
        xticks = range(0, xaxis_max + 1, xtick_gap)
        subplot.set_xticks(
            ticks=xticks
        )
        xticklabels = [
            f'{int(val / 1000)}k' for val in xticks
        ]
        subplot.set_xticklabels(
            labels=xticklabels,
            fontsize=12
        )
        subplot.tick_params(
            axis='x', which='both',
            direction='in', length=6, width=1,
            top=True, bottom=True,
            labeltop=True, labelbottom=True
        )
        subplot.grid(
            visible=True,
            which='major', axis='x',
            color='gray',
            linestyle='--', linewidth=0.3
        )
        subplot.set_xlabel(
            f'Close Value (Date: {close_date})',
            fontsize=15,
            labelpad=15
        )

        # reverse y-axis
        subplot.invert_yaxis()

        # y-axis customization
        subplot.set_ylabel(
            f'{index_type} Index',
            fontsize=15,
            labelpad=15
        )
        subplot.set_ylim(len(df), -1)

        # figure customization
        figure.suptitle(
            figure_title,
            fontsize=15,
            y=1
        )
        figure.tight_layout()
        figure.savefig(
            figure_file,
            bbox_inches='tight'
        )

        # close the figure to prevent duplicate plots from displaying
        matplotlib.pyplot.close(figure)

        return figure

    def plot_cagr_filtered_indices(
        self,
        excel_file: str,
        close_type: str,
        figure_file: str,
        threshold_cagr: float = 20.0,
        index_type: str = 'NSE Equity'
    ) -> matplotlib.figure.Figure:

        '''
        Returns a bar plot of the closing values (PRICE/TRI) of NSE indices,
        filtered by a specified threshold CAGR (%) since their launch.

        Parameters
        ----------
        excel_file : str
            Path of the input Excel file containing the data.

        close_type : str
            Type of closing value for indices, either 'PRICE' or 'TRI'.

        figure_file : str
            File Path to save the output figue.

        threshold_cagr : float, optional
            Only plot indices with a CAGR (%) higher than the specified threshold.
            Default is 20.

        index_type : str, optional
            Type of index. Default is 'NSE Equity'.

        Returns
        -------
        Figure
            A bar plot displaying closing values of NSE indices, along with
            CAGR (%), Multiplier (X), and Age (Y) since launch.
        '''

        # input DataFrame
        df = pandas.read_excel(excel_file)
        df = df[df['CAGR(%)'] >= threshold_cagr]

        # check filtered dataframe
        if len(df) == 0:
            raise Exception('Threshold values return an empty DataFrame.')
        else:
            pass

        # figure
        figure_title = (
            f'Threshold CAGR (>= {threshold_cagr} %) Since Launch: '
            f'Closing {close_type} (Bar) with CAGR (%), Multiplier (X), and Age (Y)'
        )
        figure = self._df_bar_closing(
            df=df,
            close_type=close_type,
            index_type=index_type,
            figure_title=figure_title,
            figure_file=figure_file
        )

        return figure

    def plot_top_cagr_indices(
        self,
        excel_file: str,
        close_type: str,
        figure_file: str,
        top_cagr: int = 20,
        index_type: str = 'NSE Equity'
    ) -> matplotlib.figure.Figure:

        '''
        Returns a bar plot of the closing values (PRICE/TRI) of NSE indices
        with the top CAGR (%) since their launch.

        Parameters
        ----------
        excel_file : str
            Path of the input Excel file containing the data.

        close_type : str
            Type of closing value for indices, either 'PRICE' or 'TRI'.

        figure_file : str
            File Path to save the output figue.

        top_cagr : int, optional
            The number of top indices by CAGR (%) to plot for each category.
            Default is 20.

        index_type : str, optional
            Type of index. Default is 'NSE Equity'.

        Returns
        -------
        Figure
            A bar plot displaying closing values of NSE indices, along with
            CAGR (%), Multiplier (X), and Age (Y) since launch.
        '''

        # input DataFrame
        df = pandas.read_excel(excel_file)
        df = df.head(top_cagr)

        # figure
        figure_title = (
            f'Top {top_cagr} CAGR (%) Since Launch: '
            f'Closing {close_type} (Bar) with CAGR (%), Multiplier (X), and Age (Y)'
        )
        figure = self._df_bar_closing(
            df=df,
            close_type=close_type,
            index_type=index_type,
            figure_title=figure_title,
            figure_file=figure_file
        )

        return figure

    def plot_yearwise_sip_returns(
        self,
        index: str,
        excel_file: str,
        figure_file: str,
        ytick_gap: int = 500
    ) -> matplotlib.figure.Figure:

        '''
        Generates and returns a bar plot of investments and returns for a
        monthly SIP of 1,000 Rupees over the years for a specified index.

        Parameters
        ----------
        index : str
            Name of the index.

        excel_file : str
            Path to the Excel file obtained from :meth:`BharatFinTrack.NSETRI.download_historical_daily_data`
            and :meth:`BharatFinTrack.NSETRI.update_historical_daily_data` methods.

        figure_file : str
            File Path to save the output figue.

        ytick_gap : int, optional
            Gap between two y-axis ticks. Default is 500.

        Returns
        -------
        Figure
            A bar plot displaying the investment and returns of a
            monthly SIP of 1,000 Rupees over years for the specified index.
        '''

        # check validity of input figure file path
        check_file = Core().is_valid_figure_extension(figure_file)
        if check_file:
            pass
        else:
            raise Exception('Input figure file extension is not supported.')

        # monthly investment amount
        monthly_invest = 1000

        # SIP DataFrame of index
        with tempfile.TemporaryDirectory() as tmp_dir:
            df = NSETRI().yearwise_sip_analysis(
                input_excel=excel_file,
                monthly_invest=monthly_invest,
                output_excel=os.path.join(tmp_dir, 'output.xlsx')

            )

        # figure
        fig_width = len(df) * 1.2 if len(df) >= 9 else 10
        figure = matplotlib.pyplot.figure(
            figsize=(fig_width, 10)
        )
        subplot = figure.subplots(1, 1)

        # plot bars
        xticks = pandas.Series(
            range(len(df['Year']))
        )
        bar_width = 0.4
        subplot.bar(
            x=xticks - bar_width / 2,
            height=df['Invest'] / monthly_invest,
            width=bar_width,
            label='Invest',
            color='gold'
        )
        subplot.bar(
            x=xticks + bar_width / 2,
            height=df['Value'] / monthly_invest,
            width=bar_width,
            label='Return',
            color='lightgreen'
        )
        for xt in xticks:
            multiple = df['Multiple (X)'][xt]
            xirr = df['XIRR (%)'][xt]
            subplot.annotate(
                f'{multiple:.1f}X,{xirr:.0f}%',
                xy=(xticks[xt] + bar_width / 2, df['Value'][xt] / monthly_invest),
                ha='center', va='bottom',
                fontsize=12
            )

        # x-axis customization
        subplot.set_xticks(
            ticks=xticks
        )
        xticklabels = list(
            map(
                lambda x: x[0].strftime('%d-%b-%Y') + f' ({x[1] + 1}Y)', zip(df['Start Date'], xticks)
            )
        )
        xticklabels[-1] = xticklabels[-1].replace(xticklabels[-1][12:], '\n(Launch)')
        subplot.set_xticklabels(
            labels=xticklabels,
            rotation=45,
            fontsize=12
        )
        subplot.set_xlabel(
            xlabel='SIP Start Date',
            fontsize=15
        )

        # y-axis customization
        yaxis_max = (round((df['Value'].max() / monthly_invest + 50) / ytick_gap) + 1) * ytick_gap
        subplot.set_ylim(0, yaxis_max)
        yticks = range(0, yaxis_max + 1, ytick_gap)
        subplot.set_yticks(
            ticks=yticks
        )
        yticklabels = [str(yt) + 'K' for yt in yticks]
        subplot.set_yticklabels(
            labels=yticklabels,
            fontsize=12
        )
        subplot.tick_params(
            axis='y', which='both',
            direction='in', length=6, width=1,
            left=True, right=True,
            labelleft=True, labelright=True
        )
        subplot.grid(
            visible=True,
            which='major', axis='y',
            color='gray',
            linestyle='--', linewidth=0.3
        )
        close_date = df['Close Date'].iloc[0].strftime('%d-%b-%Y')
        subplot.set_ylabel(
            ylabel=f'Amount (Date: {close_date})',
            fontsize=15
        )

        # legend
        subplot.legend(
            loc='upper left',
            fontsize=15
        )

        # figure customization
        figure_title = (
            f'{index.upper()}\n\nReturn with Multiples (X) and XIRR (%) for SIP 1000 Rupees on First Date of Each Month Over Years'
        )
        figure.suptitle(
            t=figure_title,
            fontsize=15
        )
        figure.tight_layout()
        figure.savefig(
            figure_file,
            bbox_inches='tight'
        )

        # close the figure to prevent duplicate plots from displaying
        matplotlib.pyplot.close(figure)

        return figure

    def plot_sip_index_vs_gsec(
        self,
        index: str,
        excel_file: str,
        figure_file: str,
        gsec_return: float = 8,
        ytick_gap: int = 500
    ) -> matplotlib.figure.Figure:

        '''
        Generates a bar plot comparing the returns of a specified index
        and government bonds for a monthly SIP of 1,000 Rupees over the years.'

        Parameters
        ----------
        index : str
            Name of the index.

        excel_file : str
            Path to the Excel file obtained from :meth:`BharatFinTrack.NSETRI.download_historical_daily_data`
            and :meth:`BharatFinTrack.NSETRI.update_historical_daily_data` methods.

        figure_file : str
            File Path to save the output figue.

        gsec_return : float, optional
            Expected annual return rate of government bond in percentage. Default is 8.

        ytick_gap : int, optional
            Gap between two y-axis ticks. Default is 500.

        Returns
        -------
        Figure
            A bar plot showing the return comparison between the specified index
            and government bonds for a monthly SIP of 1,000 Rupees over the years.
        '''

        # check validity of input figure file path
        check_file = Core().is_valid_figure_extension(figure_file)
        if check_file:
            pass
        else:
            raise Exception('Input figure file extension is not supported.')

        # monthly investment amount
        monthly_invest = 1000

        # SIP DataFrame of index
        with tempfile.TemporaryDirectory() as tmp_dir:
            df = NSETRI().yearwise_sip_analysis(
                input_excel=excel_file,
                monthly_invest=monthly_invest,
                output_excel=os.path.join(tmp_dir, 'output.xlsx')

            )

        # filted DataFrame
        sip_years = int(df['Year'].max())
        df = df[df['Year'] <= sip_years].reset_index(drop=True)

        # bank fixed deposit DataFrame
        bank_df = Core().sip_growth(
            invest=monthly_invest,
            frequency='monthly',
            annual_return=gsec_return,
            years=sip_years
        )

        # figure
        fig_width = len(df) * 1.2 if len(df) >= 9 else 10
        figure = matplotlib.pyplot.figure(
            figsize=(fig_width, 10)
        )
        subplot = figure.subplots(1, 1)

        # plot bars
        xticks = pandas.Series(
            range(len(df['Year']))
        )
        bar_width = 0.3
        subplot.bar(
            x=xticks - bar_width,
            height=df['Invest'] / monthly_invest,
            width=bar_width,
            label='Invest',
            color='gold'
        )
        subplot.bar(
            x=xticks,
            height=bank_df['Value'] / monthly_invest,
            width=bar_width,
            label=f'Government ({gsec_return:.1f}%)',
            color='cyan'
        )
        subplot.bar(
            x=xticks + bar_width,
            height=df['Value'] / monthly_invest,
            width=bar_width,
            label=index.upper(),
            color='lightgreen'
        )
        for xt in xticks:
            multiple = df['Multiple (X)'][xt]
            xirr = df['XIRR (%)'][xt]
            subplot.annotate(
                f'{multiple:.1f}X,{xirr:.0f}%',
                xy=(xticks[xt] + bar_width, df['Value'][xt] / monthly_invest),
                ha='center', va='bottom',
                fontsize=12
            )

        # x-axis customization
        subplot.set_xticks(
            ticks=xticks
        )
        xticklabels = list(
            map(
                lambda x: x[0].strftime('%d-%b-%Y') + f' ({x[1] + 1}Y)', zip(df['Start Date'], xticks)
            )
        )
        subplot.set_xticklabels(
            labels=xticklabels,
            rotation=45,
            fontsize=12
        )
        subplot.set_xlabel(
            xlabel='Start Date',
            fontsize=15
        )

        # y-axis customization
        yaxis_max = (round((df['Value'].max() / monthly_invest + 50) / ytick_gap) + 1) * ytick_gap
        subplot.set_ylim(0, yaxis_max)
        yticks = range(0, yaxis_max + 1, ytick_gap)
        subplot.set_yticks(
            ticks=yticks
        )
        yticklabels = [str(yt) + 'K' for yt in yticks]
        subplot.set_yticklabels(
            labels=yticklabels,
            fontsize=12
        )
        subplot.tick_params(
            axis='y', which='both',
            direction='in', length=6, width=1,
            left=True, right=True,
            labelleft=True
        )
        subplot.grid(
            visible=True,
            which='major', axis='y',
            color='gray',
            linestyle='--', linewidth=0.3
        )
        close_date = df['Close Date'].iloc[0].strftime('%d-%b-%Y')
        subplot.set_ylabel(
            ylabel=f'Amount (Date: {close_date})',
            fontsize=15
        )

        # legend
        subplot.legend(
            loc='upper left',
            fontsize=15
        )

        # figure customization
        figure_title = (
            f'{index.upper()} Return with Multiples (X) and XIRR (%)\n\nComparison Return Between Index and Government Bond for SIP 1000 Rupees on First Date of Each Month Over Years'
        )
        figure.suptitle(
            t=figure_title,
            fontsize=15
        )
        figure.tight_layout()
        figure.savefig(
            figure_file,
            bbox_inches='tight'
        )

        # close the figure to prevent duplicate plots from displaying
        matplotlib.pyplot.close(figure)

        return figure

    def plot_sip_growth_comparison_across_indices(
        self,
        indices: list[str],
        folder_path: str,
        figure_file: str,
        ytick_gap: int = 2
    ) -> matplotlib.figure.Figure:

        '''
        Generates a line plot comparing monthly SIP investment growth across multiple indices over the years.

        Parameters
        ----------
        indices : list
            A list of index names to compare in the SIP growth plot.

        folder_path : str
            Path to the directory containing Excel files with historical data for each index. Each Excel file must be
            named as '{index}.xlsx' corresponding to the index names provided in the `indices` list. These files should
            be obtained from :meth:`BharatFinTrack.NSETRI.download_historical_daily_data` or
            :meth:`BharatFinTrack.NSETRI.update_historical_daily_data`.

        figure_file : str
            File Path to save the output figue.

        ytick_gap : int, optional
            Gap between two y-axis ticks. Default is 2.

        Returns
        -------
        Figure
            A line plot figure showing cumulative SIP investment growth for each index over time.
        '''

        # check validity of input figure file path
        check_file = Core().is_valid_figure_extension(figure_file)
        if check_file:
            pass
        else:
            raise Exception('Input figure file extension is not supported.')

        # monthly investment amount
        monthly_invest = 1000

        # SIP dataframe of index
        dataframes = []
        with tempfile.TemporaryDirectory() as tmp_dir:
            for index in indices:
                index_excel = os.path.join(folder_path, f'{index}.xlsx')
                df = NSETRI().yearwise_sip_analysis(
                    input_excel=index_excel,
                    monthly_invest=monthly_invest,
                    output_excel=os.path.join(tmp_dir, 'output.xlsx')
                )
                dataframes.append(df)

        # check equal close date for all DataFrames
        close_date = dataframes[0]['Close Date'].iloc[0]
        equal_closedate = all(map(lambda df: df['Close Date'].iloc[0] == close_date, dataframes))
        if equal_closedate:
            pass
        else:
            raise Exception('Last date must be equal across all indices in the Excel files.')

        # filtered dataframes
        common_year = min(
            map(lambda df: int(df['Year'].max()), dataframes)
        )
        dataframes = list(
            map(
                lambda df: df[df['Year'] <= common_year], dataframes
            )
        )

        # figure
        figure = matplotlib.pyplot.figure(figsize=(25, 10))
        subplot = figure.subplots(1, 1)

        # plot
        colormap = matplotlib.colormaps.get_cmap('hsv')
        colors = [colormap(df / len(dataframes)) for df in range(len(dataframes))]
        for df, color, index in zip(dataframes, colors, indices):
            subplot.plot(
                df.index, df['Multiple (X)'],
                marker='o', markersize=8,
                color=color,
                label=index
            )

        # x-axis customization
        xticks = pandas.Series(
            range(common_year)
        )
        subplot.set_xticks(
            ticks=xticks
        )
        xticklabels = list(
            map(
                lambda x: x[0].strftime('%d-%b-%Y') + f' ({x[1] + 1}Y)', zip(dataframes[0]['Start Date'], xticks)
            )
        )
        subplot.set_xticklabels(
            labels=xticklabels,
            rotation=45,
            fontsize=12
        )
        subplot.set_xlabel(
            xlabel='SIP Start Date',
            fontsize=15
        )

        # y-axis customization
        growth_max = max(
            map(lambda df: int(df['Multiple (X)'].max()) + 1, dataframes)
        )
        yaxis_max = (int(growth_max / ytick_gap) + 1) * ytick_gap
        subplot.set_ylim(0, yaxis_max)
        yticks = range(0, yaxis_max + 1, ytick_gap)
        subplot.set_yticks(
            ticks=yticks
        )
        subplot.set_yticklabels(
            labels=[str(yt) for yt in yticks],
            fontsize=12
        )
        subplot.tick_params(
            axis='y', which='both',
            direction='in', length=6, width=1,
            left=True, right=True,
            labelleft=True, labelright=True
        )
        subplot.grid(
            visible=True,
            which='major', axis='y',
            color='gray',
            linestyle='--', linewidth=0.3
        )
        subplot.set_ylabel(
            ylabel=f'Multiples (X) of Investment (Date: {close_date.strftime("%d-%b-%Y")})',
            fontsize=15
        )

        # legend
        subplot.legend(
            loc='upper left',
            fontsize=12
        )

        # figure customization
        figure_title = (
            'Growth of Monthly SIP Investment on First Date of each Month Over Years'
        )
        figure.suptitle(
            t=figure_title,
            fontsize=15
        )
        figure.tight_layout()
        figure.savefig(
            figure_file,
            bbox_inches='tight'
        )

        # close the figure to prevent duplicate plots from displaying
        matplotlib.pyplot.close(figure)

        return figure



import pytest
import BharatFinTrack
import os
import tempfile
import datetime
import pandas
import matplotlib.pyplot


@pytest.fixture(scope='class')
def nse_product():

    yield BharatFinTrack.NSEProduct()


@pytest.fixture(scope='class')
def nse_index():

    yield BharatFinTrack.NSEIndex()


@pytest.fixture(scope='class')
def nse_tri():

    yield BharatFinTrack.NSETRI()


@pytest.fixture(scope='class')
def visual():

    yield BharatFinTrack.Visual()


@pytest.fixture(scope='class')
def core():

    yield BharatFinTrack.core.Core()


@pytest.fixture
def message():

    output = {
        'error_category': 'Input category "region" does not exist.',
        'error_date1': "time data '16-Sep-202' does not match format '%d-%b-%Y'",
        'error_date2': "time data '20-Se-2024' does not match format '%d-%b-%Y'",
        'error_date3': 'Start date 27-Sep-2024 cannot be later than end date 26-Sep-2024.',
        'error_excel': 'Input file extension ".xl" does not match the required ".xlsx".',
        'error_figure': 'Input figure file extension is not supported.',
        'error_folder': 'The folder path does not exist.',
        'error_index1': '"INVALID" index does not exist.',
        'error_index2': '"NIFTY50 USD" index data is not available as open-source.',
        'error_df': 'Threshold values return an empty DataFrame.',
        'error_close': 'Invalid indices close value type: TRII; must be one of [PRICE, TRI].'

    }

    return output


def test_save_dataframes_equity_indices(
    nse_product,
    message
):

    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        df = nse_product.save_dataframe_equity_index_parameters(
            excel_file=excel_file
        )
        assert len(df.index.names) == 2


def test_get_equity_indices_by_category(
    nse_product,
    message
):

    # pass test
    assert 'NIFTY 500' in nse_product.get_equity_indices_by_category('broad')
    assert 'NIFTY IT' in nse_product.get_equity_indices_by_category('sector')
    assert 'NIFTY HOUSING' in nse_product.get_equity_indices_by_category('thematic')
    assert 'NIFTY ALPHA 50' in nse_product.get_equity_indices_by_category('strategy')
    assert 'NIFTY50 USD' in nse_product.get_equity_indices_by_category('variant')

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_product.get_equity_indices_by_category('region')
    assert exc_info.value.args[0] == message['error_category']


def test_is_index_exist(
    nse_product
):

    assert nse_product.is_index_exist('NIFTY 100') is True
    assert nse_product.is_index_exist('INVALID') is False


def test_get_equity_index_base_date(
    nse_product,
    message
):

    # pass test
    assert nse_product.get_equity_index_base_date('NIFTY100 EQUAL WEIGHT') == '01-Jan-2003'
    assert nse_product.get_equity_index_base_date('NIFTY INDIA DEFENCE') == '02-Apr-2018'

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_product.get_equity_index_base_date('INVALID')
    assert exc_info.value.args[0] == message['error_index1']


def test_get_equity_index_base_value(
    nse_product,
    message
):

    # pass test
    assert nse_product.get_equity_index_base_value('NIFTY MIDCAP LIQUID 15') == 1500.0
    assert nse_product.get_equity_index_base_value('NIFTY IT') == 100.0

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_product.get_equity_index_base_value('INVALID')
    assert exc_info.value.args[0] == message['error_index1']


def test_is_index_data_open_source(
    nse_tri,
    message
):

    # pass test
    assert nse_tri.is_index_open_source('NIFTY 50') is True
    assert nse_tri.is_index_open_source('NIFTY50 USD') is False

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_tri.is_index_open_source('INVALID')
    assert exc_info.value.args[0] == message['error_index1']


def test_download_historical_daily_data(
    nse_tri,
    message
):

    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'output.xlsx')
        # error test for non open-source index
        with pytest.raises(Exception) as exc_info:
            nse_tri.download_historical_daily_data(
                index='NIFTY50 USD',
                excel_file=excel_file,
                start_date='27-Sep-2024',
                end_date='27-Sep-2024'
            )
        assert exc_info.value.args[0] == message['error_index2']
        # error test for invalid start date input
        with pytest.raises(Exception) as exc_info:
            nse_tri.download_historical_daily_data(
                index='NIFTY 50',
                excel_file=excel_file,
                start_date='16-Sep-202',
                end_date='26-Sep-2024'
            )
        assert exc_info.value.args[0] == message['error_date1']
        # error test for invalid end date input
        with pytest.raises(Exception) as exc_info:
            nse_tri.download_historical_daily_data(
                index='NIFTY 50',
                excel_file=excel_file,
                start_date='16-Sep-2024',
                end_date='20-Se-2024'
            )
        assert exc_info.value.args[0] == message['error_date2']
        # error test for strat date later than end date
        with pytest.raises(Exception) as exc_info:
            nse_tri.download_historical_daily_data(
                index='NIFTY 50',
                excel_file=excel_file,
                start_date='27-Sep-2024',
                end_date='26-Sep-2024'
            )
        assert exc_info.value.args[0] == message['error_date3']
        # pass test for start date being None
        df = nse_tri.download_historical_daily_data(
            index='NIFTY INDIA DEFENCE',
            excel_file=excel_file,
            start_date=None,
            end_date='10-Apr-2018'
        )
        assert float(df.iloc[0, -1]) == 1000.00
        # pass test for end date being None
        start_date = (datetime.date.today() - datetime.timedelta(days=7)).strftime('%d-%b-%Y')
        df = nse_tri.download_historical_daily_data(
            index='NIFTY 50',
            excel_file=excel_file,
            start_date=start_date,
            end_date=None
        )
        assert len(df) > 0


@pytest.mark.parametrize(
    'index, expected_value',
    [
        ('NIFTY INDIA SELECT 5 CORPORATE GROUPS (MAATR)', 45021.65),
        ('NIFTY INDIA RAILWAYS PSU', 4493.09),
    ]
)
def test_index_download_historical_daily_data(
    nse_tri,
    index,
    expected_value
):
    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'output.xlsx')
        df = nse_tri.download_historical_daily_data(
            index=index,
            excel_file=excel_file,
            start_date='06-Dec-2024',
            end_date='06-Dec-2024'
        )
        assert float(df.iloc[-1, -1]) == expected_value


def test_equity_index_price_download_updated_value(
    nse_index,
    message,
    capsys
):

    # error test for folder path
    with tempfile.TemporaryDirectory() as tmp_dir:
        pass
    with pytest.raises(Exception) as exc_info:
        nse_index.download_daily_summary_report(tmp_dir)
    assert exc_info.value.args[0] == message['error_folder']
    # pass test for capturing print statement
    nse_index.equity_cagr_from_launch(
        untracked_indices=True
    )
    capture_print = capsys.readouterr()
    assert 'List of untracked download indices' in capture_print.out
    assert 'List of untracked base indices' in capture_print.out

    # pass test for sorting of NSE equity indices by CAGR (%) value
    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        nse_index.sort_equity_cagr_from_launch(
            excel_file=excel_file
        )
        df = pandas.read_excel(excel_file)
        assert len(df.index.names) == 1

    # pass test for categorical sorting NSE equity indices by CAGR (%) value
    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        nse_index.category_sort_equity_cagr_from_launch(
            excel_file=excel_file
        )
        df = pandas.read_excel(excel_file, index_col=[0, 1])
        assert df.shape[1] == 9
        assert len(df.index.get_level_values('Category').unique()) <= 5


def test_equity_index_tri_closing(
    nse_tri,
    message,
    visual
):

    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        # pass test for downloading updated TRI values of NSE equity indices
        nse_tri.download_daily_summary_equity_closing(
            excel_file=excel_file,
            test_mode=True
        )
        df = pandas.read_excel(excel_file)
        assert df.shape[1] == 6
        assert df.shape[0] <= 8
        # pass test for sorting of NSE equity indices by TRI values
        output_excel = os.path.join(tmp_dir, 'tri_sort_closing_value.xlsx')
        nse_tri.sort_equity_value_from_launch(
            input_excel=excel_file,
            output_excel=output_excel
        )
        df = pandas.read_excel(output_excel)
        assert df.shape[1] == 5
        # pass test for sorting of NSE equity indices by CAGR (%) value
        output_excel = os.path.join(tmp_dir, 'tri_sort_cagr.xlsx')
        nse_tri.sort_equity_cagr_from_launch(
            input_excel=excel_file,
            output_excel=output_excel
        )
        df = pandas.read_excel(output_excel)
        assert df.shape[1] == 9
        # pass test for categorical sorting NSE equity indices by CAGR (%) value
        output_excel = os.path.join(tmp_dir, 'tri_sort_cagr_by_category.xlsx')
        nse_tri.category_sort_equity_cagr_from_launch(
            input_excel=excel_file,
            output_excel=output_excel
        )
        df = pandas.read_excel(output_excel, index_col=[0, 1])
        assert len(df.index.get_level_values('Category').unique()) <= 4
        # pass test for excess TRI CAGR(%) over PRICE CAGR(%)
        df = nse_tri.compare_cagr_over_price_from_launch(
            tri_excel=os.path.join(tmp_dir, 'tri_sort_cagr.xlsx'),
            price_excel=os.path.join(tmp_dir, 'tri_sort_cagr.xlsx'),
            output_excel=os.path.join(tmp_dir, 'compare_tri_cagr_over_price.xlsx')
        )
        assert df['Difference(%)'].unique()[0] == 0
        ###################################################################
        # PASS TEST FOR PLOTTING WITH CATEGORY
        ###################################################################
        # pass test for plotting of index closing value, grouped by category, filtered by a threshold CAGR (%) since their launch
        figure_file = os.path.join(tmp_dir, 'plot_tri_sort_cage_filtered_by_category.png')
        assert os.path.exists(figure_file) is False
        figure = visual.plot_cagr_filtered_indices_by_category(
            excel_file=output_excel,
            figure_file=figure_file,
            close_type='TRI'
        )
        assert isinstance(figure, matplotlib.pyplot.Figure) is True
        assert os.path.exists(figure_file) is True
        # pass test for plotting of index closing value, grouped by category, with the top CAGR (%) in each category since their launch.
        figure_file = os.path.join(tmp_dir, 'plot_tri_top_cagr_by_category.png')
        assert os.path.exists(figure_file) is False
        figure = visual.plot_top_cagr_indices_by_category(
            excel_file=output_excel,
            figure_file=figure_file,
            close_type='TRI'
        )
        assert isinstance(figure, matplotlib.pyplot.Figure) is True
        assert os.path.exists(figure_file) is True
        assert sum([file.endswith('.png') for file in os.listdir(tmp_dir)]) == 2
        ###################################################################
        # PASS TEST FOR PLOTTING WITHOUT CATEGORY
        ###################################################################
        # pass test for plotting of index closing value, filtered by a threshold CAGR (%) since their launch
        figure_file = os.path.join(tmp_dir, 'plot_tri_sort_filtered_cagr.png')
        assert os.path.exists(figure_file) is False
        figure = visual.plot_cagr_filtered_indices(
            excel_file=os.path.join(tmp_dir, 'tri_sort_cagr.xlsx'),
            figure_file=figure_file,
            close_type='TRI'
        )
        assert isinstance(figure, matplotlib.pyplot.Figure) is True
        assert os.path.exists(figure_file) is True
        assert sum([file.endswith('.png') for file in os.listdir(tmp_dir)]) == 3
        # pass test for plotting of index closing value with the top CAGR (%) in each category since their launch.
        figure_file = os.path.join(tmp_dir, 'plot_tri_top_cagr.png')
        assert os.path.exists(figure_file) is False
        figure = visual.plot_top_cagr_indices(
            excel_file=os.path.join(tmp_dir, 'tri_sort_cagr.xlsx'),
            figure_file=figure_file,
            close_type='TRI'
        )
        assert isinstance(figure, matplotlib.pyplot.Figure) is True
        assert os.path.exists(figure_file) is True
        assert sum([file.endswith('.png') for file in os.listdir(tmp_dir)]) == 4
        ###################################################################
        # ERROR TEST FOR PLOTTING WITH CATEGORY
        ###################################################################
        # error test for empty DataFrame from threshold values
        with pytest.raises(Exception) as exc_info:
            visual.plot_cagr_filtered_indices_by_category(
                excel_file=output_excel,
                figure_file=figure_file,
                close_type='TRI',
                threshold_cagr=100
            )
        assert exc_info.value.args[0] == message['error_df']
        # error test for invalid figure file input
        with pytest.raises(Exception) as exc_info:
            visual.plot_cagr_filtered_indices_by_category(
                excel_file=output_excel,
                figure_file=os.path.join(tmp_dir, 'error_figure_file.png'),
                close_type='TRII'
            )
        assert exc_info.value.args[0] == message['error_close']
        ###################################################################
        # ERROR TEST FOR PLOTTING WITHOUT CATEGORY
        ###################################################################
        # error test for empty DataFrame from threshold values
        with pytest.raises(Exception) as exc_info:
            visual.plot_cagr_filtered_indices(
                excel_file=output_excel,
                figure_file=figure_file,
                close_type='TRI',
                threshold_cagr=100
            )
        assert exc_info.value.args[0] == message['error_df']
        # error test for invalid input of index close type
        with pytest.raises(Exception) as exc_info:
            visual.plot_cagr_filtered_indices(
                excel_file=output_excel,
                figure_file=os.path.join(tmp_dir, 'error_figure_file.png'),
                close_type='TRII'
            )
        assert exc_info.value.args[0] == message['error_close']


def test_update_historical_daily_data(
    nse_tri
):

    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        today = datetime.date.today()
        day1_ago = today - datetime.timedelta(days=30)
        day2_ago = today - datetime.timedelta(days=15)
        # pass test for downloading daily TRI values of NSE a equity index
        nse_tri.download_historical_daily_data(
            index='NIFTY 50',
            excel_file=excel_file,
            start_date=day1_ago.strftime('%d-%b-%Y'),
            end_date=day2_ago.strftime('%d-%b-%Y')
        )
        len_df1 = len(pandas.read_excel(excel_file))
        # pass test for updating daily TRI values for the NSE equity index
        nse_tri.update_historical_daily_data(
            index='NIFTY 50',
            excel_file=excel_file
        )
        len_df2 = len(pandas.read_excel(excel_file))
        assert len_df2 > len_df1


def test_sip(
    nse_tri,
    visual
):

    with tempfile.TemporaryDirectory() as tmp_dir:
        # pass test for downloading indices historical daily data
        index = 'NIFTY 50'
        input_excel = os.path.join(tmp_dir, f'{index}.xlsx')
        nse_tri.download_historical_daily_data(
            index=index,
            excel_file=input_excel,
            start_date='15-Oct-2022',
            end_date='15-Oct-2024'
        )
        # pass test for yearwise SIP analysis
        sip_df = nse_tri.yearwise_sip_analysis(
            input_excel=input_excel,
            monthly_invest=1000,
            output_excel=os.path.join(tmp_dir, 'output.xlsx')
        )
        assert float(round(sip_df.iloc[-1, -1], 1)) == 24.0
        assert float(round(sip_df.iloc[-1, -2], 1)) == 1.3
        # pass test for SIP analysis from a fixed date
        sip_summary = nse_tri.sip_summary_from_given_date(
            excel_file=input_excel,
            start_year=2024,
            start_month=4,
            monthly_invest=1000
        )
        assert sip_summary['XIRR (%)'] == '18.4'
        # pass test for SIP analysis where given date is earlier that actual downloaded start date
        sip_summary = nse_tri.sip_summary_from_given_date(
            excel_file=input_excel,
            start_year=2021,
            start_month=4,
            monthly_invest=1000
        )
        assert sip_summary['Actual start date'] == '17-Oct-2022'
        # pass test for plotting yearwise SIP analysis
        figure_file = os.path.join(tmp_dir, f'sip_yearwise_{index}.png')
        visual.plot_yearwise_sip_returns(
            index=index,
            excel_file=input_excel,
            figure_file=figure_file,
            ytick_gap=25
        )
        assert os.path.exists(figure_file) is True
        # pass test for comparison plot of SIP for index and bank fixed depost
        figure_file = os.path.join(tmp_dir, f'sip_gsec_vs_{index}.png')
        visual.plot_sip_index_vs_gsec(
            index=index,
            excel_file=input_excel,
            figure_file=figure_file,
            gsec_return=7.5,
            ytick_gap=25
        )
        assert os.path.exists(figure_file) is True
        assert sum([file.endswith('.png') for file in os.listdir(tmp_dir)]) == 2
        # pass test for comparison plot of SIP for multiple indices
        figure_file = os.path.join(tmp_dir, 'sip_invest_growth_across_indices.png')
        visual.plot_sip_growth_comparison_across_indices(
            indices=[index],
            folder_path=tmp_dir,
            figure_file=figure_file
        )
        assert os.path.exists(figure_file) is True
        assert sum([file.endswith('.png') for file in os.listdir(tmp_dir)]) == 3
        # pass test for comparison of SIP for multiple indices
        index_1 = 'NIFTY MIDCAP150 MOMENTUM 50'
        nse_tri.download_historical_daily_data(
            index=index_1,
            excel_file=os.path.join(tmp_dir, f'{index_1}.xlsx'),
            start_date='15-Oct-2022',
            end_date='15-Oct-2024'
        )
        # SIP growth comparison
        aggregate_df = nse_tri.sip_growth_comparison_across_indices(
            indices=[index, index_1],
            folder_path=tmp_dir,
            excel_file=os.path.join(tmp_dir, 'compare_sip_growth_across_indices.xlsx')
        )
        assert len(aggregate_df.columns) == 2
        # SIP XIRR comparison
        aggregate_df = nse_tri.sip_xirr_comparison_across_indices(
            indices=[index, index_1],
            folder_path=tmp_dir,
            excel_file=os.path.join(tmp_dir, 'compare_sip_xirr_across_indices.xlsx')
        )
        assert len(aggregate_df.columns) == 2
        # error test for unequal end date of two indices
        nse_tri.download_historical_daily_data(
            index='NIFTY ALPHA 50',
            excel_file=os.path.join(tmp_dir, 'NIFTY ALPHA 50.xlsx'),
            start_date='15-Oct-2022',
            end_date='15-Sep-2024'
        )
        with pytest.raises(Exception) as exc_info:
            visual.plot_sip_growth_comparison_across_indices(
                indices=['NIFTY 50', 'NIFTY ALPHA 50'],
                folder_path=tmp_dir,
                figure_file=figure_file
            )
        assert exc_info.value.args[0] == 'Last date must be equal across all indices in the Excel files.'
        # SIP growth comparison
        with pytest.raises(Exception) as exc_info:
            nse_tri.sip_growth_comparison_across_indices(
                indices=['NIFTY 50', 'NIFTY ALPHA 50'],
                folder_path=tmp_dir,
                excel_file=os.path.join(tmp_dir, 'comare_sip_growth_across_indices.xlsx')
            )
        assert exc_info.value.args[0] == 'Last date must be equal across all indices in the Excel files.'
        # SIP XIRR comparison
        with pytest.raises(Exception) as exc_info:
            nse_tri.sip_xirr_comparison_across_indices(
                indices=['NIFTY 50', 'NIFTY ALPHA 50'],
                folder_path=tmp_dir,
                excel_file=os.path.join(tmp_dir, 'compare_sip_xirr_across_indices.xlsx')
            )
        assert exc_info.value.args[0] == 'Last date must be equal across all indices in the Excel files.'
        # error test for invalid input of year and month
        with pytest.raises(Exception) as exc_info:
            nse_tri.sip_summary_from_given_date(
                excel_file=input_excel,
                start_year=2025,
                start_month=4,
                monthly_invest=1000
            )
        assert exc_info.value.args[0] == 'Given year and month return an empty DataFrame.'


def test_error_excel(
    nse_product,
    nse_index,
    nse_tri,
    message
):

    with pytest.raises(Exception) as exc_info:
        nse_product.save_dataframe_equity_index_parameters(
            excel_file=r"C:\Users\Username\Folder\out.xl"
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_index.sort_equity_cagr_from_launch(
            excel_file='equily.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_index.category_sort_equity_cagr_from_launch(
            excel_file='equily.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.download_historical_daily_data(
            index='NIFTY 50',
            start_date='23-Sep-2024',
            end_date='27-Sep-2024',
            excel_file='NIFTY50_tri.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.download_daily_summary_equity_closing(
            excel_file='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.sort_equity_value_from_launch(
            input_excel='input.xlsx',
            output_excel='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.category_sort_equity_cagr_from_launch(
            input_excel='input.xlsx',
            output_excel='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.sort_equity_cagr_from_launch(
            input_excel='input.xlsx',
            output_excel='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.category_sort_equity_cagr_from_launch(
            input_excel='input.xlsx',
            output_excel='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.compare_cagr_over_price_from_launch(
            tri_excel='input.xlsx',
            price_excel='input.xlsx',
            output_excel='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.yearwise_sip_analysis(
            input_excel='input.xlsx',
            monthly_invest=1000,
            output_excel='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.sip_growth_comparison_across_indices(
            indices=['NIFTY 50'],
            folder_path=r"C:\Users\Username\Folder",
            excel_file='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    with pytest.raises(Exception) as exc_info:
        nse_tri.sip_xirr_comparison_across_indices(
            indices=['NIFTY 50'],
            folder_path=r"C:\Users\Username\Folder",
            excel_file='output.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']


def test_error_figure(
    visual,
    message
):

    with pytest.raises(Exception) as exc_info:
        visual._mi_df_bar_closing_with_category(
            df=pandas.DataFrame(),
            close_type='TRI',
            index_type='NSE Equity',
            figure_title='title',
            figure_file='figure_file.pn'
        )
    assert exc_info.value.args[0] == message['error_figure']

    with pytest.raises(Exception) as exc_info:
        visual._df_bar_closing(
            df=pandas.DataFrame(),
            close_type='TRI',
            index_type='NSE Equity',
            figure_title='title',
            figure_file='figure_file.pn'
        )
    assert exc_info.value.args[0] == message['error_figure']

    with pytest.raises(Exception) as exc_info:
        visual.plot_yearwise_sip_returns(
            index='NIFTY 50',
            excel_file='NIFTY 50.xlsx',
            figure_file='figure_file.pn',
            ytick_gap=250
        )
    assert exc_info.value.args[0] == message['error_figure']

    with pytest.raises(Exception) as exc_info:
        visual.plot_sip_index_vs_gsec(
            index='NIFTY 50',
            excel_file='NIFTY 50.xlsx',
            figure_file='figure_file.pn',
            gsec_return=7.5,
            ytick_gap=500
        )
    assert exc_info.value.args[0] == message['error_figure']

    with pytest.raises(Exception) as exc_info:
        visual.plot_sip_growth_comparison_across_indices(
            indices=['NIFTY 50', 'NIFTY 500'],
            folder_path=r"C:\Users\Username\Folder",
            figure_file='figure_file.pn',
            ytick_gap=2
        )
    assert exc_info.value.args[0] == message['error_figure']


def test_error_sip_growth(
    core
):

    with pytest.raises(Exception) as exc_info:
        core.sip_growth(
            invest=1000,
            frequency='monthlyy',
            annual_return=7.5,
            years=20
        )
    assert exc_info.value.args[0] == "Select a valid frequency from ['yearly', 'quarterly', 'monthly', 'weekly']"


def test_github_action(
    core
):

    assert core._github_action(
        integer=3
    ) == '3'



import os
import tempfile
import typing
import datetime
import dateutil.relativedelta
import pandas
import requests
import bs4
import matplotlib
from .nse_product import NSEProduct
from .core import Core


class NSEIndex:

    '''
    Provides functionality for downloading and analyzing
    NSE index price data (excluding dividend reinvestment).
    '''

    def download_daily_summary_report(
        self,
        folder_path: str,
        http_headers: typing.Optional[dict[str, str]] = None
    ) -> pandas.DataFrame:

        '''
        Download the daily summary for all NSE indices and save it as the
        'daily_summary_report.csv' file in the specified folder path.

        Parameters
        ----------
        folder_path: str
            Folder path to save the CSV file of the daily summary for all NSE indices.

        http_headers : dict, optional
            HTTP headers for the web request. Defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers` if not provided.

        Returns
        -------
        DataFrame
            A DataFrame containing the daily summary for all NSE indices.
        '''

        # check folder path
        if not os.path.isdir(folder_path):
            raise Exception('The folder path does not exist.')
        else:
            pass

        # web request headers
        headers = Core().default_http_headers if http_headers is None else http_headers

        # download data
        main_url = 'https://www.niftyindices.com'
        csv_url = main_url + '/reports/daily-reports'
        response = requests.get(
            url=csv_url,
            headers=headers,
            timeout=30
        )
        soup = bs4.BeautifulSoup(
            markup=response.content,
            features='html.parser'
        )
        for anchor in soup.find_all('a'):
            try:
                if anchor['href'].endswith('.csv') and anchor['id'] == 'dailysnapOneDaybefore':
                    csv_link = main_url + anchor['href']
                    response = requests.get(
                        url=csv_link,
                        headers=headers
                    )
                    download_file = os.path.join(folder_path, 'summary_index_price_closing_value.csv')
                    with open(download_file, 'wb') as download_data:
                        download_data.write(response.content)
                    output = pandas.read_csv(download_file)
                else:
                    pass
            except Exception:
                pass

        return output

    def equity_cagr_from_launch(
        self,
        http_headers: typing.Optional[dict[str, str]] = None,
        untracked_indices: bool = False
    ) -> pandas.DataFrame:

        '''
        Returns the CAGR (%) since launch for all NSE equity indices.

        Parameters
        ----------
        http_headers : dict, optional
            HTTP headers for the web request. Defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers` if not provided.

        untracked_indices : bool, optional
            Defaults to False. If True, print two lists of untracked indices
            from downloaded and based files.

        Returns
        -------
        DataFrame
            A DataFrame with the CAGR (%) since launch for all NSE equity indices.
        '''

        # downlod daily summary of NSE indices
        with tempfile.TemporaryDirectory() as tmp_dir:
            download_df = self.download_daily_summary_report(tmp_dir)

        # processing downloaded data
        date_string = datetime.datetime.strptime(
            download_df.loc[0, 'Index Date'], '%d-%m-%Y'
        )
        download_date = date_string.date()
        download_df = download_df[
            ['Index Name', 'Index Date', 'Closing Index Value']
        ]
        download_df.columns = ['Index Name', 'Close Date', 'Close Value']
        download_df['Index Name'] = download_df['Index Name'].apply(lambda x: x.upper())
        download_df['Close Date'] = download_date

        # processing base DataFrame
        base_df = NSEProduct()._dataframe_equity_index
        base_df = base_df.reset_index()
        base_df = base_df.drop(columns=['ID', 'API TRI'])
        base_df['Base Date'] = base_df['Base Date'].apply(lambda x: x.date())

        # fixing unmatched indices with base indices in the download Dataframe
        download_unmatch = {
            'NIFTY 50 FUTURES INDEX': 'NIFTY 50 FUTURES PR',
            'NIFTY 50 FUTURES TR INDEX': 'NIFTY 50 FUTURES TR',
            'NIFTY HEALTHCARE INDEX': 'NIFTY HEALTHCARE'
        }
        download_df['Index Name'] = download_df['Index Name'].apply(
            lambda x: download_unmatch.get(x, x)
        )

        # computing CAGR(%)
        cagr_df = base_df.merge(download_df)
        cagr_df['Close/Base'] = cagr_df['Close Value'] / cagr_df['Base Value']
        cagr_df['Years'] = list(
            map(
                lambda x: dateutil.relativedelta.relativedelta(download_date, x).years, cagr_df['Base Date']
            )
        )
        cagr_df['Days'] = list(
            map(
                lambda x, y: (download_date - x.replace(year=x.year + y)).days, cagr_df['Base Date'], cagr_df['Years']
            )
        )
        total_years = cagr_df['Years'] + (cagr_df['Days'] / 365)
        cagr_df['CAGR(%)'] = 100 * (pow(cagr_df['Close Value'] / cagr_df['Base Value'], 1 / total_years) - 1)

        # output
        if untracked_indices is False:
            pass
        else:
            download_index = download_df['Index Name']
            base_index = base_df['Index Name']
            untracked_download = list(download_index[~download_index.isin(base_index)])
            print(f'List of untracked download indices: {untracked_download}')
            untracked_base = list(base_index[~base_index.isin(download_index)])
            print(f'List of untracked base indices: {untracked_base}')

        return cagr_df

    def sort_equity_cagr_from_launch(
        self,
        excel_file: str,
        http_headers: typing.Optional[dict[str, str]] = None
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by CAGR (%) since launch.

        Parameters
        ----------
        excel_file : str
            Path to an Excel file to save the DataFrame.

        http_headers : dict, optional
            HTTP headers for the web request. Defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers` if not provided.

        Returns
        -------
        DataFrame
            A DataFrame sorted in descending order by CAGR (%) since launch for all NSE equity indices.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # DataFrame of CAGR(%)
        cagr_df = self.equity_cagr_from_launch(
            http_headers=http_headers,
            untracked_indices=False
        )

        # sorting DataFrame by CAGR(%)
        cagr_df = cagr_df.drop(columns=['Category'])
        cagr_df = cagr_df.sort_values(
            by=['CAGR(%)', 'Years', 'Days'],
            ascending=[False, False, False]
        )
        output = cagr_df.reset_index(drop=True)

        # saving DataFrame
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            output.to_excel(excel_writer, index=False)
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Sheet1']
            # format columns
            for col_num, col_df in enumerate(output.columns):
                if col_df == 'Index Name':
                    worksheet.set_column(col_num, col_num, 60)
                elif col_df == 'Close Value':
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0'})
                    )
                elif col_df == 'Close/Base':
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0.0'})
                    )
                elif col_df == 'CAGR(%)':
                    worksheet.set_column(
                        col_num, col_num, 15,
                        workbook.add_format({'num_format': '#,##0.00'})
                    )
                else:
                    worksheet.set_column(col_num, col_num, 15)

        return output

    def category_sort_equity_cagr_from_launch(
        self,
        excel_file: str,
        http_headers: typing.Optional[dict[str, str]] = None
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by CAGR (%)
        since launch within each index category.

        Parameters
        ----------
        excel_file : str
            Path to an Excel file to save the DataFrame.

        http_headers : dict, optional
            HTTP headers for the web request. Defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers` if not provided.

        Returns
        -------
        DataFrame
            A multi-index DataFrame sorted in descending order by CAGR (%)
            since launch within each index category.
        '''

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # DataFrame of CAGR(%)
        cagr_df = self.equity_cagr_from_launch(
            http_headers=http_headers
        )

        # Convert 'Category' column to categorical data types with a defined order
        categories = list(cagr_df['Category'].unique())
        cagr_df['Category'] = pandas.Categorical(
            cagr_df['Category'],
            categories=categories,
            ordered=True
        )

        # Sorting Dataframe
        cagr_df = cagr_df.sort_values(
            by=['Category', 'CAGR(%)', 'Years', 'Days'],
            ascending=[True, False, False, False]
        )
        dataframes = []
        for category in categories:
            category_df = cagr_df[cagr_df['Category'] == category]
            category_df = category_df.drop(columns=['Category']).reset_index(drop=True)
            dataframes.append(category_df)
        output = pandas.concat(
            dataframes,
            keys=[word.upper() for word in categories],
            names=['Category', 'ID']
        )

        # saving the DataFrame
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            output.to_excel(excel_writer, index=True)
            workbook = excel_writer.book
            worksheet = excel_writer.sheets['Sheet1']
            # number of columns for DataFrame indices
            index_cols = len(output.index.names)
            # format columns
            worksheet.set_column(0, index_cols - 1, 15)
            for col_num, col_df in enumerate(output.columns):
                if col_df == 'Index Name':
                    worksheet.set_column(index_cols + col_num, index_cols + col_num, 60)
                elif col_df == 'Close Value':
                    worksheet.set_column(
                        index_cols + col_num, index_cols + col_num, 15,
                        workbook.add_format({'num_format': '#,##0'})
                    )
                elif col_df == 'Close/Base':
                    worksheet.set_column(
                        index_cols + col_num, index_cols + col_num, 15,
                        workbook.add_format({'num_format': '#,##0.0'})
                    )
                elif col_df == 'CAGR(%)':
                    worksheet.set_column(
                        index_cols + col_num, index_cols + col_num, 15,
                        workbook.add_format({'num_format': '#,##0.00'})
                    )
                else:
                    worksheet.set_column(index_cols + col_num, index_cols + col_num, 15)
            # Dataframe colors
            get_colormap = matplotlib.colormaps.get_cmap('Pastel2')
            colors = [
                get_colormap(count / len(dataframes)) for count in range(len(dataframes))
            ]
            hex_colors = [
                '{:02X}{:02X}{:02X}'.format(*[int(num * 255) for num in color]) for color in colors
            ]
            # coloring of DataFrames
            start_col = index_cols - 1
            end_col = index_cols + len(output.columns) - 1
            start_row = 1
            for df, color in zip(dataframes, hex_colors):
                color_format = workbook.add_format({'bg_color': color})
                end_row = start_row + len(df) - 1
                worksheet.conditional_format(
                    start_row, start_col, end_row, end_col,
                    {'type': 'no_blanks', 'format': color_format}
                )
                start_row = end_row + 1

        return output



from .nse_product import NSEProduct
from .nse_index import NSEIndex
from .nse_tri import NSETRI
from .visual import Visual


__all__ = [
    'NSEProduct',
    'NSEIndex',
    'NSETRI',
    'Visual'
]


__version__ = '0.1.10'
